require 'bundler'
require 'bundler/audit/database'
require 'bundler/lockfile_parser'

require 'set'

module Bundler
  module Audit
    class Scanner

      # Represents a plain-text source
      InsecureSource = Struct.new(:source)

      # Represents a gem that is covered by an Advisory
      UnpatchedGem = Struct.new(:gem, :advisory)

      # The advisory database
      #
      # @return [Database]
      attr_reader :database

      # Project root directory
      attr_reader :root

      # The parsed `Gemfile.lock` from the project
      #
      # @return [Bundler::LockfileParser]
      attr_reader :lockfile

      #
      # Initializes a scanner.
      #
      # @param [String] root
      #   The path to the project root.
      #
      def initialize(root=Dir.pwd)
        @root     = File.expand_path(root)
        @database = Database.new
        @lockfile = LockfileParser.new(
          File.read(File.join(@root,'Gemfile.lock'))
        )
      end

      #
      # Scans the project for issues.
      #
      # @param [Hash] options
      #   Additional options.
      #
      # @option options [Array<String>] :ignore
      #   The advisories to ignore.
      #
      # @yield [result]
      #   The given block will be passed the results of the scan.
      #
      # @yieldparam [InsecureSource, UnpatchedGem] result
      #   A result from the scan.
      #
      # @return [Enumerator]
      #   If no block is given, an Enumerator will be returned.
      #
      def scan(options={})
        return enum_for(__method__,options) unless block_given?

        get_insecure_sources.each { |source| yield source }
        get_unpatched_gems(options[:ignore]).each { |gem| yield gem }

        return self
      end

    protected

      def get_insecure_sources
        insecure = []
        @lockfile.sources.each do |source|
          case source
          when Source::Git
            next unless(source.uri =~ /^(git|http):/)

            insecure << InsecureSource.new(source.uri)
          when Source::Rubygems
            source.remotes.map do |uri|
              next unless uri.scheme == 'http'

              insecure << InsecureSource.new(uri.to_s)
            end
          end
        end

        return insecure
      end

      def get_unpatched_gems(ignore)
        ignore = Set.new(ignore) # If ignore is empty the Set will contain nil,
                                 # but since we should never have a nil version
                                 # that's a non-issue.
        unpatched = []
        @lockfile.specs.each do |gem|
          @database.check_gem(gem) do |advisory|
            next if ignore.include?(advisory.id)

            unpatched << UnpatchedGem.new(gem,advisory)
          end
        end
        return unpatched
      end
    end
  end
end
